@@ -14,20 +14,18 @@
 
 The test case encapsulates the request headers and body and expected
 response headers and body. When the test is run an HTTP request is
-made using urllib3. Assertions are made against the reponse.
+made using urllib3. Assertions are made against the response.
 """
 
 from collections import OrderedDict
 import copy
 import functools
-import json
 import os
 import re
 import sys
 import time
 import unittest
-from unittest import case
-from unittest import result
+from unittest import result as unitresult
 
 import six
 from six.moves import http_cookies
@@ -36,7 +34,7 @@ import wsgi_intercept
 
 from gabbi import __version__
 from gabbi import exception
-from gabbi import json_parser
+from gabbi.handlers import base
 from gabbi import utils
 
 
@@ -49,6 +47,7 @@ REPLACERS = [
     'LOCATION',
     'COOKIE',
     'LAST_URL',
+    'URL',
     'HEADERS',
     'RESPONSE',
 ]
@@ -58,6 +57,7 @@ BASE_TEST = {
     'name': '',
     'desc': '',
     'verbose': False,
+    'cert_validate': True,
     'ssl': False,
     'redirects': False,
     'method': 'GET',
@@ -69,6 +69,7 @@ BASE_TEST = {
     'xfail': False,
     'skip': '',
     'poll': {},
+    'use_prior_test': True,
 }
 
 
@@ -80,20 +81,19 @@ def potentialFailure(func):
             try:
                 func(self)
             except Exception:
-                if hasattr(case, '_ExpectedFailure'):
-                    raise case._ExpectedFailure(sys.exc_info())
-                else:
-                    self._addExpectedFailure(self.result, sys.exc_info())
+                self._addExpectedFailure(self.result, sys.exc_info())
             else:
-                if hasattr(self, '_addUnexpectedSuccess'):
-                    self._addUnexpectedSuccess(self.result)
-                else:
-                    raise case._UnexpectedSuccess
+                self._addUnexpectedSuccess(self.result)
         else:
             func(self)
     return wrapper
 
 
+def _is_complex_type(data):
+    """If data is a list or dict return True."""
+    return isinstance(data, list) or isinstance(data, dict)
+
+
 class HTTPTestCase(unittest.TestCase):
     """Encapsulate a single HTTP request as a TestCase.
 
@@ -104,17 +104,26 @@ class HTTPTestCase(unittest.TestCase):
     tearDown are only run once.
     """
 
-    response_handlers = []
     base_test = copy.copy(BASE_TEST)
 
     def setUp(self):
+        self._fixture_cleanups = []
         if not self.has_run:
             super(HTTPTestCase, self).setUp()
+            for fixture in self.inner_fixtures:
+                f = fixture()
+                f.setUp()
+                # add fixture to front of list so we unwind them FILO in
+                # tearDown
+                self._fixture_cleanups.insert(0, f)
 
     def tearDown(self):
         if not self.has_run:
             super(HTTPTestCase, self).tearDown()
         self.has_run = True
+        # Clean up an inner fixtures.
+        for fixture in self._fixture_cleanups:
+            fixture.cleanUp()
 
     def run(self, result=None):
         """Store the current result handler on this test."""
@@ -133,26 +142,40 @@ class HTTPTestCase(unittest.TestCase):
         if self.test_data['skip']:
             self.skipTest(self.test_data['skip'])
 
-        if self.prior and not self.prior.has_run:
+        if (self.prior and not self.prior.has_run and
+                self.test_data['use_prior_test']):
             # Use a different result so we don't count this test
             # in the results.
-            self.prior.run(result.TestResult())
+            self.prior.run(unitresult.TestResult())
         self._run_test()
 
-    def replace_template(self, message):
+    def get_content_handler(self, content_type):
+        """Determine the content handler for this media type."""
+        for handler in self.content_handlers:
+            if handler.accepts(content_type):
+                return handler
+        return None
+
+    def replace_template(self, message, escape_regex=False):
         """Replace magic strings in message."""
         if isinstance(message, dict):
             for k in message:
-                message[k] = self.replace_template(message[k])
+                message[k] = self.replace_template(message[k],
+                                                   escape_regex=escape_regex)
             return message
 
+        if isinstance(message, list):
+            return [self.replace_template(line, escape_regex=escape_regex)
+                    for line in message]
+
         for replacer in REPLACERS:
             template = '$%s' % replacer
             method = '_%s_replace' % replacer.lower()
             try:
                 if template in message:
                     try:
-                        message = getattr(self, method)(message)
+                        replace = getattr(self, method)
+                        message = replace(message, escape_regex=escape_regex)
                     except (KeyError, AttributeError, ValueError) as exc:
                         raise AssertionError(
                             'unable to replace %s in %s, data unavailable: %s'
@@ -163,10 +186,13 @@ class HTTPTestCase(unittest.TestCase):
 
         return message
 
+    def load_data_file(self, filename):
+        """Read a file from the current test directory."""
+        return self._load_data_file(filename)
+
     def _assert_response(self):
         """Compare the response with expected data."""
         self._test_status(self.test_data['status'], self.response['status'])
-
         for handler in self.response_handlers:
             handler(self)
 
@@ -178,14 +204,42 @@ class HTTPTestCase(unittest.TestCase):
         value = value.encode('UTF-8')
         return value
 
-    def _environ_replace(self, message):
-        """Replace an indicator in a message with the environment value."""
+    @staticmethod
+    def _regex_replacer(replacer, escape_regex):
+        """Wrap a replacer function to escape return values in a regex."""
+        if escape_regex:
+            @functools.wraps(replacer)
+            def replace(match):
+                return re.escape(replacer(match))
+
+            return replace
+        else:
+            return replacer
+
+    def _environ_replace(self, message, escape_regex=False):
+        """Replace an indicator in a message with the environment value.
+
+        If value can be a number, cast it as such. If value is a form of
+        "null", "true", or "false" cast it to None, True, False.
+        """
         value = re.sub(self._replacer_regex('ENVIRON'),
-                       self._environ_replacer, message)
-        if value == "False":
+                       self._regex_replacer(self._environ_replacer,
+                                            escape_regex),
+                       message)
+        try:
+            if '.' in value:
+                value = float(value)
+            else:
+                value = int(value)
+            return value
+        except ValueError:
+            pass
+        if value.lower() == "false":
             return False
-        if value == "True":
+        if value.lower() == "true":
             return True
+        if value.lower() == "null":
+            return None
         return value
 
     @staticmethod
@@ -197,76 +251,116 @@ class HTTPTestCase(unittest.TestCase):
         environ_name = match.group('arg')
         return os.environ[environ_name]
 
-    @staticmethod
-    def extract_json_path_value(data, path):
-        """Extract the value at JSON Path path from the data.
-
-        The input data is a Python datastructure, not a JSON string.
-        """
-        path_expr = json_parser.parse(path)
-        matches = [match.value for match in path_expr.find(data)]
-        if matches:
-            if len(matches) > 1:
-                return matches
-            else:
-                return matches[0]
-        else:
-            raise ValueError(
-                "JSONPath '%s' failed to match on data: '%s'" % (path, data))
-
-    def _cookie_replace(self, message):
+    def _cookie_replace(self, message, escape_regex=False):
         """Replace $COOKIE in a message.
 
         With cookie data from set-cookie in the prior request.
         """
-        response_cookies = self.prior.response['set-cookie']
+        return re.sub(self._simple_replacer_regex('COOKIE'),
+                      self._regex_replacer(self._cookie_replacer,
+                                           escape_regex),
+                      message)
+
+    def _cookie_replacer(self, match):
+        """Replace a regex match with the cookie of a previous response."""
+        case = match.group('case')
+        if case:
+            referred_case = self.history[case]
+        else:
+            referred_case = self.prior
+        response_cookies = referred_case.response['set-cookie']
         cookies = http_cookies.SimpleCookie()
         cookies.load(response_cookies)
         cookie_string = cookies.output(attrs=[], header='', sep=',').strip()
-        return message.replace('$COOKIE', cookie_string)
+        return cookie_string
 
-    def _headers_replace(self, message):
-        """Replace a header indicator in a message with that headers value from
-        the prior request.
+    def _headers_replace(self, message, escape_regex=False):
+        """Replace a header indicator in a message.
+
+        Replace it with the header's value from the prior request.
         """
         return re.sub(self._replacer_regex('HEADERS'),
-                      self._header_replacer, message)
+                      self._regex_replacer(self._header_replacer,
+                                           escape_regex),
+                      message)
 
     def _header_replacer(self, match):
         """Replace a regex match with the value of a prior header."""
         header_key = match.group('arg')
-        return self.prior.response[header_key.lower()]
+        case = match.group('case')
+        if case:
+            referred_case = self.history[case]
+        else:
+            referred_case = self.prior
+        return referred_case.response[header_key.lower()]
 
-    def _json_replacer(self, match):
-        """Replace a regex match with the value of a JSON Path."""
-        path = match.group('arg')
-        return str(self.extract_json_path_value(self.prior.json_data, path))
-
-    def _last_url_replace(self, message):
+    def _last_url_replace(self, message, escape_regex=False):
         """Replace $LAST_URL in a message.
 
         With the URL used in the prior request.
         """
-        return message.replace('$LAST_URL', self.prior.url)
+        last_url = self.prior.url
+        if escape_regex:
+            last_url = re.escape(last_url)
+        return message.replace('$LAST_URL', last_url)
 
-    def _location_replace(self, message):
+    def _url_replace(self, message, escape_regex=False):
+        """Replace $URL in a message.
+
+        With the URL used in a previous request.
+        """
+        return re.sub(self._simple_replacer_regex('URL'),
+                      self._regex_replacer(self._url_replacer,
+                                           escape_regex),
+                      message)
+
+    def _url_replacer(self, match):
+        """Replace a regex match with the value of a previous url."""
+        case = match.group('case')
+        if case:
+            referred_case = self.history[case]
+        else:
+            referred_case = self.prior
+        return referred_case.url
+
+    def _location_replace(self, message, escape_regex=False):
         """Replace $LOCATION in a message.
 
-        With the location header from the prior request.
+        With the location header from a previous request.
         """
-        return message.replace('$LOCATION', self.prior.location)
+        return re.sub(self._simple_replacer_regex('LOCATION'),
+                      self._regex_replacer(self._location_replacer,
+                                           escape_regex),
+                      message)
+
+    def _location_replacer(self, match):
+        """Replace a regex match with the value of a previous location."""
+        case = match.group('case')
+        if case:
+            referred_case = self.history[case]
+        else:
+            referred_case = self.prior
+        return referred_case.location
 
     def _load_data_file(self, filename):
         """Read a file from the current test directory."""
-        path = os.path.join(self.test_directory, os.path.basename(filename))
+        path = os.path.join(self.test_directory, filename)
+        has_dir_traversal = os.path.relpath(
+            path, start=self.test_directory).startswith(os.pardir)
+        if has_dir_traversal:
+            raise ValueError(
+                'Attempted loading of data file outside test directory: %s'
+                % filename)
         with open(path, mode='rb') as data_file:
             return data_file.read()
 
-    def _netloc_replace(self, message):
+    def _netloc_replace(self, message, escape_regex=False):
         """Replace $NETLOC with the current host and port."""
         netloc = self.netloc
         if self.prefix:
             netloc = '%s%s' % (netloc, self.prefix)
+        if escape_regex:
+            netloc = re.escape(netloc)
         return message.replace('$NETLOC', netloc)
 
     def _parse_url(self, url):
@@ -299,15 +393,57 @@ class HTTPTestCase(unittest.TestCase):
         return urlparse.urlunsplit((parsed_url.scheme, parsed_url.netloc,
                                     parsed_url.path, query_string, ''))
 
+    _history_regex = (
+        r"(?:\$HISTORY\[(?P<quote1>['\"])(?P<case>.+?)(?P=quote1)\]\.)??"
+    )
+
     @staticmethod
     def _replacer_regex(key):
         """Compose a regular expression for test template variables."""
-        return r"\$%s\[(?P<quote>['\"])(?P<arg>.+?)(?P=quote)\]" % key
+        case = HTTPTestCase._history_regex
+        return r"%s\$%s\[(?P<quote>['\"])(?P<arg>.+?)(?P=quote)\]" % (
+            case, key)
 
-    def _response_replace(self, message):
-        """Replace a JSON Path from the prior request with a value."""
-        return re.sub(self._replacer_regex('RESPONSE'),
-                      self._json_replacer, message)
+    @staticmethod
+    def _simple_replacer_regex(key):
+        """Compose a regular expression for simple variable replacement."""
+        case = HTTPTestCase._history_regex
+        return r"%s\$%s" % (case, key)
+
+    def _response_replace(self, message, escape_regex=False):
+        """Replace a content path with the value from a previous response.
+
+        If the match would replace the entire message, then don't cast it
+        to a string.
+        """
+        regex = self._replacer_regex('RESPONSE')
+        match = re.match('^%s$' % regex, message)
+        if match:
+            return self._response_replacer(match, preserve=True)
+        return re.sub(regex,
+                      self._regex_replacer(self._response_replacer,
+                                           escape_regex),
+                      message)
+
+    def _response_replacer(self, match, preserve=False):
+        """Replace a regex match with the value from a previous response."""
+        response_path = match.group('arg')
+        case = match.group('case')
+        if case:
+            referred_case = self.history[case]
+        else:
+            referred_case = self.prior
+        replacer_class = self.get_content_handler(
+            referred_case.response.get('content-type'))
+        # If no handler can be found use the null replacer,
+        # which returns "foo" when "$RESPONSE['foo']".
+        replacer_class = replacer_class or base.ContentHandler
+        result = replacer_class.replacer(
+            referred_case.response_data, response_path)
+        if preserve:
+            return result
+        else:
+            return six.text_type(result)
 
     def _run_request(self, url, method, headers, body, redirect=False):
         """Run the http request and decode output.
@@ -341,14 +477,29 @@ class HTTPTestCase(unittest.TestCase):
         # Decode and store response
         decoded_output = utils.decode_response_content(response, content)
         self.content_type = response.get('content-type', '').lower()
-        if (decoded_output and
-                ('application/json' in self.content_type or
-                 '+json' in self.content_type)):
-            self.json_data = json.loads(decoded_output)
+        loader_class = self.get_content_handler(self.content_type)
+        if decoded_output and loader_class:
+            # save structured response data
+            self.response_data = loader_class.loads(decoded_output)
         else:
-            self.json_data = None
+            self.response_data = None
         self.output = decoded_output
 
+    def _replace_headers_template(self, test_name, headers):
+        replaced_headers = {}
+
+        try:
+            for name in headers:
+                replaced_name = self.replace_template(name)
+                replaced_headers[replaced_name] = self.replace_template(
+                    headers[name]
+                )
+        except TypeError as exc:
+            raise exception.GabbiFormatError(
+                'malformed headers in test %s: %s' % (test_name, exc))
+
+        return replaced_headers
+
     def _run_test(self):
         """Make an HTTP request and compare the response with expectations."""
         test = self.test_data
@@ -358,25 +509,32 @@ class HTTPTestCase(unittest.TestCase):
         self.url = base_url
         full_url = self._parse_url(base_url)
 
+        # Replace variables in headers with variable values. This includes both
+        # in the header key and the header value.
+        test['request_headers'] = self._replace_headers_template(
+            test['name'], test['request_headers'])
+        test['response_headers'] = self._replace_headers_template(
+            test['name'], test['response_headers'])
+
         method = test['method'].upper()
         headers = test['request_headers']
-        for name in headers:
-            try:
-                headers[name] = self.replace_template(headers[name])
-            except TypeError as exc:
-                raise exception.GabbiFormatError(
-                    'malformed headers in test %s: %s' % (test['name'], exc))
 
-        if test['data'] is not '':
+        if test['data'] != '':
             body = self._test_data_to_string(
                 test['data'],
                 utils.extract_content_type(headers, default='')[0])
         else:
             body = ''
 
+        # ensure body is bytes, encoding as UTF-8 because that's
+        # what we do here
+        if isinstance(body, six.text_type):
+            body = body.encode('UTF-8')
+
         if test['poll']:
-            count = test['poll'].get('count', 1)
-            delay = test['poll'].get('delay', 1)
+            count = int(float(self.replace_template(
+                test['poll'].get('count', 1))))
+            delay = float(self.replace_template(test['poll'].get('delay', 1)))
             failure = None
             while count:
                 try:
@@ -398,29 +556,46 @@ class HTTPTestCase(unittest.TestCase):
                               redirect=test['redirects'])
             self._assert_response()
 
-    def _scheme_replace(self, message):
+    def _scheme_replace(self, message, escape_regex=False):
         """Replace $SCHEME with the current protocol."""
-        return message.replace('$SCHEME', self.scheme)
+        scheme = re.escape(self.scheme) if escape_regex else self.scheme
+        return message.replace('$SCHEME', scheme)
 
     def _test_data_to_string(self, data, content_type):
         """Turn the request data into a string.
 
         If the data is not binary, replace template strings.
+
+        If the result of the template handling is not a string,
+        run the result through the dumper.
         """
-        if isinstance(data, str):
-            if data.startswith('<@'):
-                info = self._load_data_file(data.replace('<@', '', 1))
+        dumper_class = self.get_content_handler(content_type)
+        if not _is_complex_type(data):
+            if isinstance(data, six.string_types) and data.startswith('<@'):
+                info = self.load_data_file(data.replace('<@', '', 1))
                 if utils.not_binary(content_type):
-                    try:
-                        info = str(info, 'UTF-8')
-                    except TypeError:
-                        info = info.encode('UTF-8')
-                    data = info
+                    data = six.text_type(info, 'UTF-8')
                 else:
+                    # Return early we are binary content
                     return info
         else:
-            data = json.dumps(data)
-        return self.replace_template(data)
+            # We have a complex data structure, try to dump it.
+            if dumper_class:
+                data = self.replace_template(data)
+                data = dumper_class.dumps(data, test=self)
+            else:
+                raise ValueError(
+                    'unable to process data to %s' % content_type)
+
+        data = self.replace_template(data)
+
+        # If the result after template handling is not a string, dump
+        # it if there is a suitable dumper.
+        if dumper_class and not isinstance(data, six.string_types):
+            # If there are errors dumping we want them to raise to the
+            # test harness.
+            data = dumper_class.dumps(data, test=self)
+        return data
 
     def _test_status(self, expected_status, observed_status):
         """Confirm we got the expected status.
@@ -471,9 +646,13 @@ class HTTPTestCase(unittest.TestCase):
             if expected in iterable:
                 return
 
-            if self.json_data:
-                full_response = json.dumps(self.json_data, indent=2,
-                                           separators=(',', ': '))
+            if self.response_data:
+                dumper_class = self.get_content_handler(self.content_type)
+                if dumper_class:
+                    full_response = dumper_class.dumps(self.response_data,
+                                                       pretty=True, test=self)
+                else:
+                    full_response = self.output
             else:
                 full_response = self.output
 
