@@ -18,6 +18,8 @@ The key piece of code is :meth:`test_suite_from_dict`. It produces a
 """
 
 import copy
+import functools
+import unittest
 
 from gabbi import case
 from gabbi.exception import GabbiFormatError
@@ -79,12 +81,21 @@ class TestMaker(object):
         self._validate_keys(test, test_name)
 
         http_class = httpclient.get_http(verbose=test['verbose'],
-                                         caption=test['name'])
+                                         caption=test['name'],
+                                         cert_validate=test['cert_validate'])
         if prior_test:
             history = prior_test.history
         else:
             history = {}
 
+        test_method_name = 'test_request'
+        test_method = getattr(case.HTTPTestCase, test_method_name)
+
+        @unittest.skipIf(self.host == '', 'No host configured')
+        @functools.wraps(test_method)
+        def do_test(*args, **kwargs):
+            return test_method(*args, **kwargs)
+
         # Use metaclasses to build a class of the necessary type
         # and name with relevant arguments.
         klass = TestBuilder(test_name, (case.HTTPTestCase,),
@@ -101,6 +112,7 @@ class TestMaker(object):
                              'prefix': self.prefix,
                              'prior': prior_test,
                              'history': history,
+                             test_method_name: do_test,
                              })
         # We've been asked to, make this test class think it comes
         # from a different module.
@@ -207,10 +219,13 @@ def test_suite_from_dict(loader, test_base_name, suite_dict, test_directory,
 
     # Merge global with per-suite defaults
     default_test_dict = copy.deepcopy(case.HTTPTestCase.base_test)
+    seen_keys = set()
     for handler in handlers:
-        default_test_dict.update(handler.test_base)
+        default_test_dict.update(copy.deepcopy(handler.test_base))
         if handler.response_handler:
-            response_handlers.append(handler.response_handler)
+            if handler.test_key_suffix not in seen_keys:
+                response_handlers.append(handler.response_handler)
+                seen_keys.add(handler.test_key_suffix)
         if handler.content_handler:
             content_handlers.append(handler.content_handler)
 
